#!/usr/bin/env ruby

require 'monitor'
require 'resolv'
require 'socket'
require 'timeout'

# Copyright (c) 2007 FlexiGuided GmbH, Berlin
#
# Author: Jan Behrens
#
# Website: http://www.flexiguided.de/publications.udat.en.html
#
# -----
#
# This module provides data structures to represent nodes of UDAT
# documents, a data format offering a generic basis for data storage and
# transmission, while being both easily readable by humans and machines.
# The format is comparable to formats like XML or YAML, but due to its
# simplicity is much more easy to parse.
#
# UDAT documents can be generated by converting an Object to a Udat::Node,
# by calling Object#to_udat, and then encoding them by calling
# Udat::Node#encode.
#
# UDAT documents can be parsed by calling String#parse_udat.
#
# There are methods providing useful IO functions for UDAT documents:
# IO#read_udat, IO#write_udat, Udat::Node#rpc and Udat.run_rpc_server.
module Udat

  module_function

  # Error raised, when a fetched Udat::Node doesn't have expected
  # properties.
  class UdatPropertyMismatch < StandardError; end
  # Error raised, when a fetched Udat::Node doesn't have the expected tag
  # (i.e. type).
  class UdatTagMismatch < UdatPropertyMismatch; end
  # Error raised, when a fetched Udat::Node is a collection instead of a
  # scalar or vice versa.
  class UdatTypeMismatch < UdatPropertyMismatch; end
  # Error raised, when an Udat::Node could not be found by
  # Udat::Collection#fetch.
  class UdatIndexError < IndexError; end
  # UDAT parsing error
  class ParseError < StandardError; end

  # Special argument used as an argument default, do not use it directly.
  AnyTag = Object.new

  # Returns an escaped version of a string, where backslashes are
  # preceding certain reserved characters.
  def escape_string(string)
    string.to_s.gsub /([<>\[\]|~\\])/, "\\\\\\1"
  end

  # The abstract class to represent any UDAT data is a Udat::Node.
  # Udat::Node's basically consist of a tag and the content. The tag can be
  # a String or be nil. The content is either a scalar value stored as a
  # String, or an ordered/unordered collection of values (where each value
  # can optionally have a key associated with it). Keys and values of
  # collections are always Udat::Node's too.
  #
  # Udat::Node's can most easily be constructed from other ruby objects by
  # using the Object#to_udat method.
  #
  # By calling the method Udat::Node#encode, the Node and its children are
  # encoded in a both easily human readable and easily machine readable
  # format. The output can be later parsed by String#parse_udat or
  # IO#read_udat.
  class Node

    public

    include MonitorMixin

    private_class_method :new
    # Creates a new Udat::Node with a given tag (which may be nil). This
    # method is private, Udat::Node#initialize is only called through
    # supercalls.
    def initialize(tag)
      super()
      self.tag = tag
    end

    # Tag (i.e. type of the content).
    attr_reader :tag
    # Sets the tag (i.e. type of the content).
    def tag=(tag)
      synchronize do
        @tag = tag ? tag.to_s.dup.freeze : nil
      end
      return tag
    end

    # Returns true, if the Udat::Node represents a scalar value.
    def scalar?
      false
    end
    # Returns true, if the Udat::Node represents a collection.
    def collection?
      false
    end

    # Returns self, if the tag is matching the argument,
    # otherwise an Udat::UdatTagMismatch exception is raised.
    def require_tag(tag)
      synchronize do
        if self.tag == tag.to_s
          return self
        else
          raise UdatTagMismatch, "UDAT tag mismatch."
        end
      end
    end
    # Returns self, if the Udat::Node represents a scalar,
    # otherwise an Udat::UdatTypeMismatch exception is raised.
    # If an optional tag is given as an argument,
    # it is checked by calling Udat::Node#require_tag.
    def require_scalar(tag = AnyTag)
      if scalar?
        if tag == AnyTag
          return self
        else
          return require_tag(tag)
        end
      elsif collection?
        raise UdatTypeMismatch,
          "UDAT collection found where a scalar was expected."
      else
        raise "Internal error in UDAT library."
      end
    end
    # Returns self, if the Udat::Node represents a collection,
    # otherwise an Udat::UdatTypeMismatch exception is raised.
    # If an optional tag is given as an argument,
    # it is checked by calling Udat::Node#require_tag.
    def require_collection(tag = AnyTag)
      if collection?
        if tag == AnyTag
          return self
        else
          return require_tag(tag)
        end
      elsif scalar?
        raise UdatTypeMismatch,
          "UDAT scalar found where a collection was expected."
      else
        raise "Internal error in UDAT library."
      end
    end

    # Returns the encoded form of the content as a string.
    # This method is used by Udat::Node#encode_without_brackets.
    def encode_content
    end
    protected :encode_content
    # Returns the data (including it's tag) encoded in the UDAT format.
    # The result is not enclosed by square brackets, so not intended to be
    # used directly externally.
    def encode_without_brackets
      synchronize do
        if tag
          return "#{Udat::escape_string tag}|#{encode_content}"
        else
          return encode_content
        end
      end
    end
    protected :encode_without_brackets
    # Returns the data (including it's tag) encoded in the UDAT format and
    # enclosed in square brackets.
    #
    # Note: The UDAT format doesn't contain information, whether contained
    # collections are ordered or unordered. This information is lost during
    # the encoding process, and has to be restored in an application
    # specific way, if neccessary.
    def encode
      "[#{encode_without_brackets}]"
    end
    # Alias for Udat::Node#encode, will be removed in future versions.
    def encode_document
      encode
    end
    # Here the method does the same as Udat::Node#encode,
    # but this method is overwritten in Udat::Scalar!
    def to_s
      encode
    end
    # Does the same as Udat::Node#encode,
    # but preceds the result with "udat".
    def inspect
      "udat#{self.encode}"
    end

    # Returns self, or a duplicate of self with a different tag set, if an
    # argument is passed.
    def to_udat(tag = AnyTag)
      if tag == AnyTag
        return self
      else
        obj = nil
        synchronize do
          obj = self.dup
        end
        obj.tag = tag
        return obj
      end
    end

    # Returns a hash key used by ruby's Hash'es.
    def hash
      to_s.hash
    end
    # Returns true, if class, tag and content (including it's order in case
    # of a Udat::Collection) are matching another object.
    def eql?(other)
      self.class == other.class and
        self.tag == other.tag and
        self.to_s == other.to_s
    end
    # Same as Udat::Node#eql?, but behaves differently in Udat::Collection.
    def ==(other)
      self.eql? other
    end

    # Internal parsing function.
    def self.parse_intern(mode, end_char, &input_reader)
      string = ""
      tag = nil
      collection = nil
      key = nil
      redo_char = false
      while true
        if redo_char
          redo_char = false
        else
          char = input_reader.call
        end
        case char
        when nil
          if end_char.nil? and not key
            break
          else
            raise EOFError, "Unexpected end of UDAT input."
          end
        when "\\"
          char = input_reader.call
          if char.nil?
            raise EOFError, "Unexpected end of UDAT input."
          end
          if char =~ /([<>\[\]|~\\])/
            string << char if string
          elsif char == "\n"
          elsif char == "\r"
            char = input_reader.call
            unless char == "\n"
              redo_char = true
              redo
            end
          elsif char == "#"
            boundary = ""
            while true
              char = input_reader.call
              if char.nil?
                raise EOFError, "Unexpected end of UDAT input."
              end
              if char == "#"
                break
              else
                boundary << char
              end
            end
            while true
              char = input_reader.call
              if char.nil?
                raise EOFError,
                  "Unexpected end of UDAT input " <<
                  "before ending boundary."
              end
              if char == "#"
                boundary_end = ""
                while true
                  char = input_reader.call
                  if char.nil?
                    raise EOFError,
                      "Unexpected end of UDAT input " <<
                      "before ending boundary."
                  end
                  if char == "#"
                    if boundary_end == boundary
                      break
                    else
                      string << "#" << boundary_end if string
                      boundary_end = ""
                      redo
                    end
                  else
                    boundary_end << char
                  end
                end
                break
              else
                string << char if string
              end
            end
          elsif char == "$"
            length = 0
            while true
              char = input_reader.call
              if char.nil?
                raise EOFError, "Unexpected end of UDAT input."
              elsif char =~ /[0-9]/
                length *= 10
                length += char.to_i
              elsif char == "$"
                break
              else
                raise ParseError,
                  "Illegal character in length information of UDAT input."
              end
            end
            length.times do
              char = input_reader.call
              if char.nil?
                raise EOFError,
                  "Unexpected end of UDAT input in binary part."
              end
              string << char if string
            end
          else
            raise ParseError, "Unknown escape sequence in UDAT input."
          end
        when "|"
          if tag or not string
            raise ParseError, "Unexpected pipe symbol in UDAT input."
          end
          tag = string
          string = ""
        when "~"
          collection ||= Collection.new(tag, [])
          string = nil
        when "<"
          if key
            raise ParseError,
              "Unexpected opening angle bracket in UDAT input."
          end
          key = parse_intern(:normal, ">", &input_reader)
          string = nil
        when ">"
          if end_char == ">" and not key
            break
          else
            raise ParseError,
              "Unexpected closing angle bracket in UDAT input."
          end
        when "["
          value = parse_intern(:normal, "]", &input_reader)
          return value if mode == :one_value
          collection ||= Collection.new(tag, [])
          if key
            collection.append_pair(key, value)
            key = nil
          else
            collection.append_value(value)
          end
          string = nil
        when "]"
          if end_char == "]" and not key
            break
          else
            raise ParseError,
              "Unexpected closing square bracket in UDAT input."
          end
        else
          string << char if string
        end
      end
      return nil if mode == :one_value
      return collection ? collection : string.to_udat(tag)
    end
    private_class_method :parse_intern

    # Parses a given UDAT document string and returns a structure of
    # Udat::Node's.
    #
    # Note: When parsing UDAT data, no information is gained, whether
    # collections are ordered or unordered. After parsing, all collections
    # will be marked as unordered, unless changed later by the application.
    def self.parse(input)
      input = input.to_s
      pos = 0
      begin
        collection = (
          parse_intern(:normal, nil) do
            char = input[pos, 1]
            pos += 1
            next char.empty? ? nil : char
          end
        )
        unless collection.collection? and not collection.empty?
          raise ParseError, "No valid UDAT object found."
        end
        return collection.first
      rescue EOFError
        raise ParseError, $!.message
      end
    end
    # Alias for Udat::Node.parse, will be removed in future versions.
    def self.parse_document(input)
      parse(input)
    end

    # Reads an encoded Udat::Node from a stream.
    def self.read_from_stream(io)
      return (
        parse_intern(:one_value, nil) do
          io.read(1)
        end
      )
    end
    # Encodes an object in the UDAT format by calling Object#to_udat
    # followed by Udat::Node#encode, and writes it to a stream.
    # Returns self.
    def write_to_stream(io)
      io << self.to_udat.encode
      return self
    end

    # Sends the object in encoded form through a TCP connection to a given
    # host and port, or if no port is supplied to the host and port
    # specified in a DNS SRV record "_udat-rpc._tcp.<domain>". Returns the
    # UDAT object received as a reply. As this method can block for an
    # unlimited amount of time, it might be useful to enclose the call in a
    # Timeout.timeout call.
    def rpc(domain, port = nil)
      domain = domain.to_str
      port = port.to_int if port
      socket_args = nil
      result = nil
      if port.nil?
        Resolv::DNS.open do |dns|
          srv_rr = dns.getresource(
            "_udat-rpc._tcp.#{domain}",
            Resolv::DNS::Resource::IN::SRV
          )
          socket_args = [srv_rr.target.to_s, srv_rr.port.to_i]
        end
      else
        socket_args = [domain, port]
      end
      socket = nil
      begin
        socket = TCPSocket.new(*socket_args)
        socket.write_udat(self)
        socket.close_write
        result = socket.read_udat
      ensure
        socket.close if socket
      end
      unless result
        raise EOFError, "End of file before reading UDAT RPC result."
      end
      return result
    end

  end

  # Class of Udat::Node's holding an ordered or unordered collection of
  # values (where each value can optionally have a key associated with it).
  # Keys and values of collections are always Udat::Node's too.
  # Udat::Collection's can be interpreted as maps, arrays, sets or any
  # other non-scalar data structure.
  #
  # Note: Keys and values are always implicitly casted by all instance
  # methods of this class to Udat::Node's using Object#to_udat.
  class Collection < Node

    public

    # Internally used wrapper class for Udat::Node's, to lookup
    # Udat::Collection's in Hash'es, while ignoring the order of their
    # content.
    class UnorderedWrapper
      def initialize(content)
        @content = content
      end
      attr_accessor :content
      def hash
        if content.kind_of? Collection
          hash = 0
          content.key_value_pairs.each do |key, value|
            hash += key.hash
            hash += value.hash
          end
          return hash
        else
          return content.hash
        end
      end
      def eql?(other)
        return false unless self.class == other.class
        unless self.content.kind_of? Node and
            other.content.kind_of? Node
          return self.content.eql?(other.content)
        end
        if self.content.class == other.content.class
            self.content.tag == other.content.tag
          if self.content.kind_of? Collection
            own_pairs = self.content.key_value_pairs
            other_pairs = other.content.key_value_pairs
            own_pairs.each do |own_pair|
              catch :found do
                other_pairs.each_index do |index|
                  if own_pair.eql? other_pairs[index]
                    other_pairs.delete_at index
                    throw :found
                  end
                end
                return false
              end
            end
            return true
          else
            return self.content.eql?(other.content)
          end
        else
          return false
        end
      end
      def ==(other)
        self.eql? other
      end
    end

    public_class_method :new
    # Creates a new Udat::Collection with a given tag (which may be nil)
    # and a given content, represented by a nested Array structure. It is
    # not recommended to use this method. Use Object#to_udat instead.
    def initialize(tag, content)
      super tag
      @ordered = true
      @entries = []
      require_index_hashes
      content.to_ary.each { |entry| self << entry }
    end

    # Returns true.
    def collection?
      true
    end

    # Deletes the internal index hashes.
    # They are automatically reconstructed once a lookup is done.
    # This method should be called by the user, if contained Udat objects
    # have changed, AFTER being added to this collection (similar to
    # Hash#rehash).
    # Returns self.
    def rehash
      synchronize do
        @key_indicies = nil
        @value_indicies = nil
        @unordered_key_indicies = nil
        @unordered_value_indicies = nil
        @all_unordered_key_indicies = nil
        @all_unordered_value_indicies = nil
      end
      return self
    end

    # Returns true, if the internal index hashes are not existent.
    def index_hashes_void?
      synchronize do
        return @key_indicies.nil?
      end
    end
    private :index_hashes_void?
    # Registers a key value pair in the index.
    # (The key argument may be nil for values without keys.)
    def add_to_index_hashes(index, key, value)
      unless index_hashes_void?
        @key_indicies[key] ||= []
        @key_indicies[key] << index
        @value_indicies[value] ||= []
        @value_indicies[value] << index
        if key.kind_of? Collection and key.unordered?
          @unordered_key_indicies[UnorderedWrapper.new(key)] ||= []
          @unordered_key_indicies[UnorderedWrapper.new(key)] << index
        end
        if value.kind_of? Collection and value.unordered?
          @unordered_value_indicies[UnorderedWrapper.new(value)] ||= []
          @unordered_value_indicies[UnorderedWrapper.new(value)] << index
        end
        @all_unordered_key_indicies[UnorderedWrapper.new(key)] ||= []
        @all_unordered_key_indicies[UnorderedWrapper.new(key)] << index
        @all_unordered_value_indicies[UnorderedWrapper.new(value)] ||= []
        @all_unordered_value_indicies[UnorderedWrapper.new(value)] << index
      end
      return self
    end
    private :add_to_index_hashes
    # Creates index hashes, if they are not existent.
    def require_index_hashes
      synchronize do
        if index_hashes_void?
          @key_indicies = {}
          @value_indicies = {}
          @unordered_key_indicies = {}
          @unordered_value_indicies = {}
          @all_unordered_key_indicies = {}
          @all_unordered_value_indicies = {}
          @entries.each_index do |index|
            key_value_pair = @entries[index]
            key = key_value_pair[0]
            value = key_value_pair[1]
            add_to_index_hashes(index, key, value) unless key.nil?
          end
        end
      end
      return self
    end
    private :require_index_hashes

    # Empties the collection. Returns self.
    def clear
      synchronize do
        @entries.clear
        rehash
        require_index_hashes
      end
      return self
    end
    # Deletes the entry at the given numeric index.
    # Returns the value, or nil if the index is out of bounds.
    def delete_at(index)
      index = index.to_int
      synchronize do
        key_value_pair = @entries.delete_at(index)
        rehash
        return key_value_pair ? key_value_pair[1] : nil
      end
    end
    # Deletes the first entry and returns its value.
    def shift
      delete_at(0)
    end
    # Deletes the last entry and returns its value.
    def pop
      delete_at(-1)
    end
    # Deletes all entries with a given key. Returns self.
    def delete_key(key)
      key = key.to_udat
      synchronize do
        @entries.delete_if { |e_key, e_value| e_key == key }
        rehash
      end
      return self
    end
    # Deletes all entries with a given value. Returns self.
    def delete_value(value)
      value = value.to_udat
      synchronize do
        @entries.delete_if { |e_key, e_value| e_value == value }
        rehash
      end
      return self
    end

    # Appends a new key with a new value. Returns self.
    def append_pair(key, value)
      key = key.to_udat
      value = value.to_udat
      synchronize do
        index = @entries.length
        @entries << [key, value]
        add_to_index_hashes(index, key, value)
      end
      return self
    end
    # Appends a new value. Returns self.
    def append_value(value)
      value = value.to_udat
      synchronize do
        index = @entries.length
        @entries << [nil, value]
        add_to_index_hashes(index, nil, value)
      end
      return self
    end
    # Deletes all key/value pairs where the given key matches,
    # and appends a new key/value pair. Returns self.
    def set_value_for_key(key, value)
      delete_key(key)
      append_pair(key, value)
      return self
    end
    # Deletes all values or key/value pairs where the given value matches,
    # and appends a new key/value pair. Returns self.
    def set_key_for_value(key, value)
      delete_value(value)
      append_pair(key, value)
      return self
    end

    # Behaves differently depending on the type of the argument. If the
    # argument is an Array with 2 elements, a new key/value pair is added,
    # with the key being the first element of the array, and the value
    # being the second argument of the array. Arrays with any other number
    # of elements cause an error. If the argument is not an array, then it
    # is added as a value without a key.
    def <<(value)
      if value.kind_of? Array
        unless value.length == 2
          raise ArgumentError,
            "#{value.length} array elements found where 2 were expected."
        end
        return append_pair(value[0], value[1])
      else
        return append_value(value)
      end
    end
    # Same as Udat::Collection#set_value_for_key.
    def []=(key, value)
      set_value_for_key(key, value)
    end

    # Returns the value at a numeric index, or nil, if the index is out of
    # bounds.
    def value_by_index(index)
      index = index.to_int
      synchronize do
        entry = @entries[index]
        return entry ? entry[1] : nil
      end
    end
    # Returns the key at a numeric index, or nil, if the index is out of
    # bounds or at the given index there is only a value without key.
    def key_by_index(index)
      index = index.to_int
      synchronize do
        entry = @entries[index]
        return entry ? entry[0] : nil
      end
    end
    # Returns an Array of values having no key.
    def values_without_key
      synchronize do
        require_index_hashes
        return (
          (@key_indicies[nil] || []).collect { |index| @entries[index][1] }
        )
      end
    end
    # Returns an Array of values having a given key.
    def values_by_key(key)
      key = key.to_udat
      synchronize do
        require_index_hashes
        if key.kind_of? Collection and key.unordered?
          indicies = @all_unordered_key_indicies[
            UnorderedWrapper.new(key)
          ] || []
        else
          indicies = (@key_indicies[key] || []) + (
            @unordered_key_indicies[UnorderedWrapper.new(key)] || []
          )
          indicies.uniq!
          indicies.sort!
        end
        return indicies.collect { |index| @entries[index][1] }
      end
    end
    # Returns an Array of keys having a given value. The Array may contain
    # nil's for values having no key.
    def keys_by_value(value)
      value = value.to_udat
      synchronize do
        require_index_hashes
        if value.kind_of? Collection and value.unordered?
          indicies = @all_unordered_value_indicies[
            UnorderedWrapper.new(value)
          ] || []
        else
          indicies = (@value_indicies[value] || []) + (
            @unordered_value_indicies[UnorderedWrapper.new(value)] || []
          )
          indicies.uniq!
          indicies.sort!
        end
        return indicies.collect { |index| @entries[index][0] }
      end
    end
    # Returns the last value without key.
    def value_without_key
      values_without_key.last
    end
    # Returns the last value having a given key.
    def value_by_key(key)
      values_by_key(key).last
    end
    # Returns the last key having a given value.
    def key_by_value(value)
      keys_by_value(value).last
    end
    # Returns true, if the value is contained in the collection.
    def include?(value)
      value = value.to_udat
      synchronize do
        require_index_hashes
        if value.kind_of? Collection and value.unordered?
          return (@all_unordered_value_indicies[
            UnorderedWrapper.new(value)
          ] || []).empty? ? false : true
        else
          return ((@value_indicies[value] || []) + (
            @unordered_value_indicies[UnorderedWrapper.new(value)] || []
          )).empty? ? false : true
        end
      end
    end

    # Behaves differently depending on the type of the argument.
    # If the argument is an Integer, the method behaves same as
    # Udat::Collection#value_by_index. If the argument is nil, the method
    # behaves like Udat::Collection#value_without_key. Otherwise the method
    # behaves same as Udat::Collection#value_by_key.
    def [](key)
      if key.nil?
        return value_without_key
      elsif key.kind_of? Integer
        return value_by_index(key)
      else
        return value_by_key(key)
      end
    end
    # Same as Udat::Collection#[], but raises an Udat::UdatIndexError, if
    # no value was found. If an additional second argument is passed, an
    # error will also be raised if the tag (i.e. type) of the value is not
    # equal to the second argument.
    def fetch(key, tag = AnyTag)
      value = self[key]
      unless value
        raise UdatIndexError, "UDAT node not found."
      end
      value.require_tag(tag) unless tag == AnyTag
      return value
    end
    # Same as Udat::Collection#fetch, but raises an error, if the value is
    # not an Udat::Scalar.
    def fetch_scalar(key, tag = AnyTag)
      value = fetch(key)
      value.require_scalar
      value.require_tag(tag) unless tag == AnyTag
      return value
    end
    # Same as Udat::Collection#fetch, but raises an error, if the value is
    # not an Udat::Collection.
    def fetch_collection(key, tag = AnyTag)
      value = fetch(key)
      value.require_collection
      value.require_tag(tag) unless tag == AnyTag
      return value
    end
    # Returns the first value of the collection, or nil if empty.
    def first
      value_by_index(0)
    end
    # Returns the last value of the collection, or nil if empty.
    def last
      value_by_index(-1)
    end

    # Returns the number of values in the collection.
    def length
      synchronize do
        @entries.length
      end
    end
    alias size length
    # Returns true, if the collection is empty.
    def empty?
      length == 0
    end
    # Calls a given block for each numeric index.
    def each_index
      synchronize do
        (0...length).each { |i| yield i }
      end
      return self
    end

    # Returns an Array containing the key/value pairs each as an Array of
    # size 2. If there is no key, the first element of the sub Array is
    # nil.
    def key_value_pairs
      synchronize do
        return @entries.collect { |entry| entry.dup }
      end
    end
    # Returns an Array containing all keys of the collection.
    def keys
      keys = nil
      synchronize do
        keys = @entries.collect { |key, value| key }
      end
      keys.compact!
      return keys
    end
    # Returns an Array containing all values of the collection.
    def values
      synchronize do
        return @entries.collect { |key, value| value }
      end
    end

    # Returns a hash, where each key is mapped to the respective value.
    def to_hash
      hash = {}
      synchronize do
        @entries.each do |key, value|
          next if key.nil?
          hash[key] = value unless hash.has_key? key
        end
      end
      return hash
    end
    # Same as Udat::Collection#values.
    def to_ary
      values
    end
    alias to_a values

    # Appends the values (and corresponding keys) of another
    # Udat::Collection and returns self.
    def concat(other)
      synchronize do
        other.key_value_pairs.each do |key, value|
          if key.nil?
            append_value(value)
          else
            append_pair(key, value)
          end
        end
      end
      return self
    end
    # Replaces the values (and corresponding keys) with the values/keys of
    # another Udat::Collection and returns self.
    def replace(other)
      synchronize do
        clear
        concat(other)
      end
      return self
    end

    # Same as Udat::Collection#ordered?.
    def ordered
      synchronize do
        return @ordered
      end
    end
    # Returns true, if the order of the contents of this collection is
    # significant for comparisons.
    def ordered?
      self.ordered
    end
    # Returns true, if the order of the contents of this collection is to
    # be ignored for comparisons.
    def unordered?
      not self.ordered
    end
    # If set to false, the order of the contents of this collection will be
    # ignored for comparisons.
    def ordered=(ordered)
      synchronize do
        if ordered == false
          @ordered = false
        else
          @ordered = true
        end
      end
    end
    # Same as Udat::Collection#ordered = false, but returns self.
    def unordered!
      self.ordered = false
      return self
    end
    # Same as Udat::Collection#ordered = true, but returns self.
    def ordered!
      self.ordered = true
      return self
    end

    # Returns the encoded form of the content as a string.
    # This method is used by Udat::Node#encode_without_brackets.
    def encode_content
      synchronize do
        return (
          @entries.empty? ? "~" :
          @entries.collect do |key, value|
            if key
              "<#{key.encode_without_brackets}>" <<
              "[#{value.encode_without_brackets}]"
            else
              "[#{value.encode_without_brackets}]"
            end
          end.join
        )
      end
    end
    # Same as Udat::Node#inspect, but in case of unordered collections it
    # prepends "udat-unordered" instead of just "udat".
    def inspect
      if ordered?
        return super
      else
        "udat-unordered#{self.encode_document}"
      end
    end

    # Returns true, if class, tag and content are matching another object.
    # The order of the content is only significant for comparison, if both
    # objects have the Udat::Collection#ordered attribute set to true.
    def ==(other)
      return false unless self.class == other.class
      if self.unordered? or other.unordered?
        return UnorderedWrapper.new(self) ==
          UnorderedWrapper.new(other)
      else
        return super
      end
    end

  end


  # Class of Udat::Node's holding scalar values (stored as a String).
  class Scalar < Node

    public

    public_class_method :new
    # Creates a new Udat::Scalar with a given tag (which may be nil) and a
    # given content, which is transformed to a string (via Object#to_s).
    # It is not recommended to use this method. Use Object#to_udat instead.
    def initialize(tag, content)
      super tag
      self.content = content
    end

    # Returns true.
    def scalar?
      true
    end

    # Content of the scalar (a String).
    attr_reader :content
    # Sets the content. The content is casted to a string.
    def content=(content)
      @content = content.to_s
    end

    # Same as Udat::Scalar#content.
    def to_s
      content
    end
    # Returns the content casted to an Integer.
    def to_i
      content.to_i
    end
    # Returns the content casted to a Float.
    def to_f
      content.to_f
    end

    # Returns true, if the (String representation of the) content is empty.
    def empty?
      self.to_s.empty?
    end

    # Returns the encoded form of the content as a string. In this case
    # this is simply an escaped version of the String.
    def encode_content
      Udat::escape_string(self.to_s)
    end

  end

  # Waits in an endless loop for incoming connections and runs a new thread
  # for each incoming connection, reads and parses UDAT objects, passes
  # them each to a given block (by calling yield) and writes the result of
  # the called block in form of an encoded UDAT object back to the socket.
  # If the argument "listen_on" is an Integer, it is used as a TCP port
  # number, if the argument is a String, the port being listened on will
  # be determined by a DNS look-up of an _udat-rpc._tcp SRV record. In both
  # cases the TCP socket is bound to all interfaces. Alternatively
  # "listen_on" may be a non Integer object, supplying an "accept" method
  # like TCPServer or UNIXServer, which returns a socket for each incoming
  # connection.
  def run_rpc_server(listen_on, timeout = 180)
    unless block_given?
      raise ArgumentError, "No block given for Udat::run_rpc_server call."
    end
    if listen_on.kind_of? String
      Resolv::DNS.open do |dns|
        srv_rr = dns.getresource(
          "_udat-rpc._tcp.#{listen_on}",
          Resolv::DNS::Resource::IN::SRV
        )
        listen_on = srv_rr.port.to_i
      end
    end
    new_server = nil
    begin
      if listen_on.kind_of? Integer
        new_server = TCPServer.new(listen_on)
        server = new_server
      elsif listen_on.respond_to? :accept
        server = listen_on
      elsif listen_on.respond_to? :to_int
        new_server = TCPServer.new(listen_on.to_int)
        server = new_server
      else
        raise ArgumentError,
          "Argument \"listen_on\" passed to Udat::run_rpc_server " <<
          "is not an integer and has no \"accept\" or \"to_int\" method."
      end
      while true
        Thread.new(server.accept) do |socket|
          begin
            while true
              query = nil
              begin
                Timeout.timeout(timeout) do
                  begin
                    query = socket.read_udat
                  rescue EOFError, ParseError
                    break
                  end
                end
              rescue Timeout::Error
              end
              break unless query
              socket.write_udat(yield(query))
            end
          ensure
            socket.close
          end
        end
      end
    ensure
      new_server.close if server
    end
  end

  # Module containing some demonstration functions.
  module Demo

    module_function

    # A string containing an UDAT example document.
    ExampleDocument = <<-'END_OF_EXAMPLE'
    [sample config v1.0|

      <network> [
        <max connections> [256]
        <reverse lookups> [yes]
      ]

      <locale> [
        <language> [de]    german language
        <timezone> [CET]   and CET timezone
        Comments can appear almost anywhere.
      ]

      <access control> [
        <allowed> [
          [user|martin]
          [user|max]
          [group|admins]
          [ip4network|
            <address> [192.168.0.0]
            <netmask> [255.255.255.0]
          ]
        ]
        <blocked> [~]   The tilde symbol denotes an empty collection.
      ]

      <misc> [
        <symbol for homedir> [\~]   Not an empty collection but the scalar
                                    value "~".
      ]

      <address mappings> [

        <email|
          <user>   [jan.behrens]
          <domain> [flexiguided.de]
        >
        [mbox|jan]

        <email|
          <user>   [glob|*]
          <domain> [flexiguided.de]
        >
        [mbox|catchall]

      ]

      <logging> [
        <verbosity>   [high]
    \## <verbosity>   [debug] ##   Uncomment this for debug output.
        <destination> [file|/var/log/sample.log]
      ]

    ]
    END_OF_EXAMPLE

    # Runs a demo server, not for production use.
    def run_rpc_demo_server(port, timeout = 180)
      Udat::run_rpc_server(port, timeout) do |query|
        begin
          case query.tag
          when "echo request"
            query.tag = "echo reply"
            next query
          when "calculation request"
            begin
              query.require_collection
              operands = query.fetch_collection("operands")
              operand1 = operands.fetch_scalar(0).to_f
              operand2 = operands.fetch_scalar(1).to_f
            end
            operand1 = operands[0].to_f
            operand2 = operands[1].to_f
            operator = query.fetch_scalar("operator")
            case operator
            when "+".to_udat
              next (operand1 + operand2).to_udat("calculation result")
            when "-".to_udat
              next (operand1 - operand2).to_udat("calculation result")
            when "*".to_udat
              next (operand1 * operand2).to_udat("calculation result")
            when "/".to_udat
              next (operand1.quo operand2).to_udat("calculation result")
            else
              next([["message", "Unknown operator."]].to_udat("error")
              )
            end
          when "time request"
            next Time.now.strftime("%Y-%m-%d %H:%M:%S %Z").to_udat('time')
          else
            raise Udat::UdatTagMismatch, "Unknown UDAT tag."
          end
        rescue Udat::UdatPropertyMismatch, UdatIndexError
          next [["message", $!.message]].to_udat("error")
        end
      end
    end

    # Method to test the RPC mechanism, not for production use.
    def rpc_selftest(port = 10330)
      a = rand(100) + 1
      b = rand(100) + 1
      demo_server_thread = nil
      begin
        demo_server_thread = Thread.new do
          Udat::Demo::run_rpc_demo_server(port)
        end
        sleep 0.5
        udat_result = nil
        Timeout.timeout(15) do
          udat_result = {'operands' => [a, b], 'operator' => '+'}.
                        to_udat("calculation request").
                        rpc('localhost', port)
        end
        if udat_result.tag == "error"
          raise "RPC call returned with an error: " <<
            udat_result.fetch_scalar("message").to_s
        end
        result = udat_result.require_scalar("calculation result").to_f
      ensure
        demo_server_thread.kill
      end
      unless result == a + b
        raise "Calculation error during UDAT RPC selftest."
      end
      return "Test done: #{a} + #{b} == #{result}"
    end

  end

end


class Object
  public
  # Casts the Object to an Udat::Scalar object, optionally with a given
  # tag. Unless overwritten by sub classes, only the string representation
  # (as returned by Object#to_s) will be stored as content in the
  # Udat::Scalar object.
  def to_udat(tag = nil)
    Udat::Scalar.new(tag, self)
  end
end

class Array
  public
  # Casts the Array to an Udat::Collection object, optionally with a given
  # tag. The Udat::Collection object is marked to be ordered. Each element
  # of the array may be an Array of size 2, in which case the 2 entries are
  # used as key and value. If an element of the Array is not an Array, it
  # will be used as value without an associated key.
  def to_udat(tag = nil)
    Udat::Collection.new(tag, self)
  end
end

class Hash
  public
  # Casts the Hash to an Udat::Collection object, optionally with a given
  # tag. The Udat::Collection object is marked to be unordered.
  def to_udat(tag = nil)
    Udat::Collection.new(tag, self.to_a).unordered!
  end
end

class String
  public
  # Calls Udat::Node.parse(self).
  def parse_udat
    Udat::Node.parse(self)
  end
  # Alias for Udat::Node.parse, will be removed in future versions.
  def parse_udat_document
    Udat::Node.parse_document(self)
  end
end

class IO
  public
  # Calls Udat::Node.read_from_stream(self).
  def read_udat
    Udat::Node.read_from_stream(self)
  end
  # Calls Udat::Node#write_to_stream(self), after converting the object to
  # an Udat::Node by calling Object#to_udat. Returns self.
  def write_udat(object)
    object.to_udat.write_to_stream(self)
    return self
  end
end

